/*
 *  RVanalysisGlobalPermutation.cpp
 *  LogitReg
 *
 *  Created by Youna Hu on 1/6/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include "RVanalysisGlobalPermutation.h"

// simple intializer
void RVanalysisGlobalPermutation::Initializer(){
	nPermutation = 0;
	nGenes = 0;
}

void RVanalysisGlobalPermutation::Initializer(RareVariantAnalysis &rvAnalysisBaseValue, int timesPermutation){
	eachPermutation = rvAnalysisBaseValue;
	nGenes = 0;
	nPermutation = timesPermutation;
	pQuantilesPermutation.Dimension(nPermutation, 2);
}

void RVanalysisGlobalPermutation::Initializer(RareVariantAnalysis &rvAnalysisBaseValue, Matrix &xAllGenes, 
																							Vector &yObs, int timesPermutation){
	nSample = yObs.Length();
	eachPermutation = rvAnalysisBaseValue;
	yPermuted.Dimension(nSample);
	nGenes = 0;
	nPermutation = timesPermutation;
	pGenesEachPermute.Dimension(nPermutation);
	pQuantilesPermutation.Dimension(nPermutation, 2);	
}


void RVanalysisGlobalPermutation::Initializer(RareVariantAnalysis &rvAnalysisBaseValue, Matrix &xAllGenes, 
																							Vector &yObs, int timesPermutation, 
																							Matrix &positionMatrix, 
																							const std::string &geneListFileName){
	nSample = yObs.Length();
	eachPermutation = rvAnalysisBaseValue;
	yPermuted.Dimension(nSample);
	nGenes = 0;
	nPermutation = timesPermutation;
	pGenesEachPermute.Dimension(nPermutation);
	pQuantilesPermutation.Dimension(nPermutation, 2);	
	// update the gene infomation
	RVanalysisUpdateGeneInfo(positionMatrix, geneListFileName);
}

// global permutation
RVanalysisGlobalPermutation::RVanalysisGlobalPermutation()
{
	Initializer();
}

RVanalysisGlobalPermutation::RVanalysisGlobalPermutation(RareVariantAnalysis &rvAnalysisBaseValue, Matrix &xAllGenes, 
																												 Vector &yObs, int timesPermutation){
	Initializer(rvAnalysisBaseValue, xAllGenes, yObs, timesPermutation);
}


RVanalysisGlobalPermutation::RVanalysisGlobalPermutation(RareVariantAnalysis &rvAnalysisBaseValue, Matrix &xAllGenes, 
																												 Vector &yObs, int timesPermutation, Matrix &positionMatrix, 
																												 const std::string &geneListFileName){
	Initializer(rvAnalysisBaseValue, xAllGenes, yObs, timesPermutation, positionMatrix, geneListFileName);
}

// destructor - do nothing
RVanalysisGlobalPermutation::~RVanalysisGlobalPermutation(){
}

// update nGenes, vec_idx_start, vec_idx_end using the Subset012byGene class
void RVanalysisGlobalPermutation::RVanalysisUpdateGeneInfo(Matrix &positionMatrix, const std::string & geneListFileName){
	// subset to obtain gene level covariate matrices
	// Read in the position matrix of the rare variants
	PositionFile posfile(positionMatrix);
	GeneFile genefile(geneListFileName);
	int iStartPos, iEndPos, totalNumGenes = genefile.GetGeneCount();
	int startGeneCount = 0;
//	totalNumGenes = 60;
	
//	PrintVecIndices();
	for (int i = startGeneCount; i < totalNumGenes; i ++) {
		const Gene & gene = genefile.GetGene(i);
		if (posfile.FindRange(gene.iChromosome, gene.iStart, gene.iEnd, iStartPos, iEndPos)) {
	//		std::cout << "Gene " << i << " " <<  gene.strName << " : " << iStartPos << ", " << iEndPos << std::endl;
				if (iStartPos != posfile.TOO_SMALL & iStartPos != posfile.TOO_BIG & iStartPos != posfile.NOTHING_IN_RANGE & iStartPos != posfile.SEARCH_FAIL
					& iEndPos != posfile.TOO_SMALL & iEndPos != posfile.TOO_BIG & iEndPos != posfile.NOTHING_IN_RANGE & iEndPos != posfile.SEARCH_FAIL) {
					nGenes = nGenes + 1;
					// update the vec_geneInfo vector
					GeneOutINFO geneNew;
					geneNew.basicGeneInfo = gene;
					geneNew.basicGeneInfo.iStart = iStartPos;
					geneNew.basicGeneInfo.iEnd = iEndPos;// remember that iStartPos and iEndPos are relative indices in the rv matrix
					geneNew.numCase1 = -1;
					geneNew.numCtrl1 = -1;
					vec_geneInfo.push_back(geneNew);
			}
			// printf("nGenes = %d\n",nGenes);
		}			
	}
	// allocate pGeneEacPermute
	pGenesEachPermute.Dimension(nGenes);
	idxLower = (int)((double)nGenes*0.05); // 5% quantile
	idxUpper = (int)((double)nGenes*0.95); // 95% quantile
	statObs.Dimension(nGenes);
	statEachPermute.Dimension(nGenes);
	// allocate and initialize eachGeneSameScore
	eachGeneSameScore.resize(nGenes);
	for (int i = 0; i < nGenes; i ++) {
		eachGeneSameScore[i] = false;
	}
	pFinalEachGene.Dimension(nGenes, 0);
	pGenesAsy.Dimension(nGenes, 0);
}

void RVanalysisGlobalPermutation::RVanalysisUpdateEachGene(Matrix & xAllGenes, int idxGeneIn){ // note that the idxGene is not the index of gene at the gene file
	
	nSample = xAllGenes.rows;
	nEachGeneStart = vec_geneInfo[idxGeneIn].basicGeneInfo.iStart; //vec_idx_start[idxGeneIn];
	nEachGeneSize = vec_geneInfo[idxGeneIn].basicGeneInfo.iEnd - nEachGeneStart + 1; //vec_idx_end[idxGeneIn] - nEachGeneStart + 1;
	xEachGene.Dimension(nSample, nEachGeneSize);
	// update the genotype matrix for this particular gene
		for (int i = 0; i < nSample; i ++) {
			for (int j = 0; j < nEachGeneSize; j++) {
				xEachGene[i][j] = xAllGenes[i][j+nEachGeneStart];
			}
		}
}

// global permutation steps: 1. subset matrix,  2. permute y, 3. calcualte p value for 
// need to intialize with timesPermutation -- just keep one copy of the Initializer (TODO)
// if we have covariate, it can be used in initializer for eachPermutation as well
void RVanalysisGlobalPermutation::GlobalPermutation(Matrix &xAllGenes, Matrix &positionMatrix, const std::string & geneListFileName, Vector &yObs, std::string choice, Vector &userWeight){
	// record the gene information -- starting and ending index at the big matrix
	RVanalysisUpdateGeneInfo(positionMatrix,geneListFileName);
	// the index for y
	for (int i = 0; i < nSample; i ++) {
		orderPermuted.push_back(i);
	}
	
	// permute y and each time, record the 5% and 95% of the pvalues of all genes
	for (int i = 0; i < nPermutation; i ++) {
		gsl_ran_shuffle(g_r, &(orderPermuted[0]), nSample, sizeof(int));
		// update the phenotype at this permutation order
		for (int j = 0; j < nSample; j ++) { // update the permuted phenotype
			yPermuted[j] = yObs[orderPermuted[j]];
		}
		
		// go through all genes
		for (int l = 0; l < nGenes ; l++) { // using the same phentoype for each gene
			// updat the xEachGene
			if (!eachGeneSameScore[l]) {
				RVanalysisUpdateEachGene(xAllGenes, l);
				// need to update the memembers in eachPermutation
				eachPermutation.Initializer(xEachGene,yPermuted);
				eachPermutation.RVregression(choice,userWeight); // note: Do need need to regress those gene with same score
				pGenesEachPermute[l] = eachPermutation.p;
				eachGeneSameScore[l] = eachPermutation.sameScore;
			}
		}
		
		// sort the p's
		pGenesEachPermute.Sort();
		pQuantilesPermutation[i][0] = pGenesEachPermute[idxLower];
		pQuantilesPermutation[i][1] = pGenesEachPermute[idxUpper];	
	}  
}


void RVanalysisGlobalPermutation::PvaluesGeneAsy(Matrix &xAllGenes, Vector &yObs, std::string choice, Vector &userWeight){
	// use a vector to keep the original order of the index for y
	/*for (int i = 0; i < nSample; i ++) {
		orderPermuted.push_back(i);
	} */
	
	// update the observed statistics for each gene
	for (int i = 0; i < nGenes; i ++) {
		// updat the xEachGene
		if (!eachGeneSameScore[i]){
			RVanalysisUpdateEachGene(xAllGenes, i);
			eachPermutation.Initializer(xEachGene,yObs);
			// regress on the observed phenotypes
			eachPermutation.RVregression(choice,userWeight);
			// update the numCase1 and numCtrl1 in vec_GeneInfo
			vec_geneInfo[i].numCase1 = eachPermutation.twoBytwoMat.Get01();
			vec_geneInfo[i].numCtrl1 = eachPermutation.twoBytwoMat.Get11();
			// update the p values 
			pGenesAsy[i] = eachPermutation.p;
			// update the same score indicator
			eachGeneSameScore[i] = eachPermutation.sameScore;
		}
	}
}	

void RVanalysisGlobalPermutation::PvaluesGeneAsyWithCov(Matrix &xAllGenes, Matrix &xCovIn, 
																												Vector &yObs, std::string choice, Vector &userWeight){
	// update the observed statistics for each gene
	for (int i = 0; i < nGenes; i ++) {
		// update the xEachGene
		RVanalysisUpdateEachGene(xAllGenes, i);
		eachPermutation.Initializer(xEachGene,xCovIn, yObs);
		// regress on the observed phenotypes
		eachPermutation.RVregressionWithCov(choice,userWeight);
		// update the p values -- this is useful for the genes with uniform rv scores across all subjects
		pGenesAsy[i] = eachPermutation.p; 
		// update the numCase1 and numCtrl1 in vec_GeneInfo
		vec_geneInfo[i].numCase1 = eachPermutation.twoBytwoMat.Get01();
		vec_geneInfo[i].numCtrl1 = eachPermutation.twoBytwoMat.Get11();		
		// update the same score indicator
		eachGeneSameScore[i] = eachPermutation.sameScore;
	}	
}


void RVanalysisGlobalPermutation::PvaluesUpdateStep(int i, double statPermuted, double statObserved){ 
	// note: Here I could easily write a counter but the advantage of using the Pvalue() in RVanalysis is that there are options suh as 
	// if the test is two sided and if the extreme statistics are bigger or smaller ones
	// note: Need to make sure nPermute is 1 when using the Pvalue()
	if(eachPermutation.nPermute != 1){
		printf("Error: Need to only have one permutation for RVanalysis");
		return;
	}
	eachPermutation.statistics = statObserved;
	eachPermutation.statPermute[0] = statPermuted;
	eachPermutation.Pvalue(false, false); 
	pFinalEachGene[i] += eachPermutation.p;
//	printf("stat obs = %f, stat permuted = %f, p = %f\n",statObserved,statPermuted,eachPermutation.p);
}

void RVanalysisGlobalPermutation::PvaluesFinalizeStep(){
	// work on pFinalEachGene -- only need to work on those without same score for all subjects
	for (int i = 0; i < nGenes; i ++) {
		if (!eachGeneSameScore[i]) {
			pFinalEachGene[i] = pFinalEachGene[i]/nPermutation;
		}
	}
}

void RVanalysisGlobalPermutation::PvaluesGenePermutation(Matrix &xAllGenes, Vector &yObs, 
																												 std::string choice, Vector &userWeight){
	// use a vector to keep the original order of the index for y
	for (int i = 0; i < nSample; i ++) {
		orderPermuted.push_back(i);
	}
	
	// update the observed statistics for each gene
	for (int i = 0; i < nGenes; i ++) {
		//		printf("same score = %s\n",eachGeneSameScore[i]?"true":"false");
		// update the xEachGene
		RVanalysisUpdateEachGene(xAllGenes, i);
		eachPermutation.Initializer(xEachGene,yObs);
		// regress on the observed phenotypes
		eachPermutation.RVregPermuteObs(choice,userWeight);
		// update the observed statitics
		statObs[i] = eachPermutation.statistics;
		// update the p values -- this is useful for the genes with uniform rv scores across all subjects
		pFinalEachGene[i] = eachPermutation.p; 
		// update the same score indicator
		eachGeneSameScore[i] = eachPermutation.sameScore;
		// for those that desires permuation (!sameScore) , we should intialize their p values to be 0
		if (!eachGeneSameScore[i]) {
			pFinalEachGene[i] = 0;
		}
		// update the numCase1 and numCtrl1 in vec_GeneInfo
		vec_geneInfo[i].numCase1 = eachPermutation.twoBytwoMat.Get01();
		vec_geneInfo[i].numCtrl1 = eachPermutation.twoBytwoMat.Get11();
	}
		
	// permute y and each time, update those that does not have the same rare variant scores
	for (int i = 0; i < nPermutation; i ++) {
		// shuffle the order
		gsl_ran_shuffle(g_r, &(orderPermuted[0]), nSample, sizeof(int));
		
		// update the phenotype at this permutation order
		for (int j = 0; j < nSample; j ++) { // update the permuted phenotype
			yPermuted[j] = yObs[orderPermuted[j]];
		}
		
		// go through all genes
		for (int l = 0; l < nGenes ; l++) { // using the same phentoype for each gene
			if (!eachGeneSameScore[l]) {
				// updat the xEachGene
				RVanalysisUpdateEachGene(xAllGenes, l);
				// need to update the memembers in eachPermutation
				eachPermutation.Initializer(xEachGene,yPermuted,1); 
				eachPermutation.RVregPermuteObs(choice,userWeight); // Note: Don't use RVPermuteEach() since the xGenoCov needs to be updated here every time
				statEachPermute[l] = eachPermutation.statistics;
				pGenesEachPermute[l] = eachPermutation.p;
				// update the p values at this step 
				PvaluesUpdateStep(l,statEachPermute[l],statObs[l]); 
			}
		}
		
		// sort the p's
		pGenesEachPermute.Sort();
		pQuantilesPermutation[i][0] = pGenesEachPermute[idxLower];
		pQuantilesPermutation[i][1] = pGenesEachPermute[idxUpper];	
	}
	
  // after the permutation is done, need to finalize the p values
	 PvaluesFinalizeStep();
}


void RVanalysisGlobalPermutation::PvaluesGenePermutationWithCov(Matrix &xAllGenes, Matrix &xCovIn, 
																																Vector &yObs, std::string choice, 
																																Vector &userWeight){
	// use a vector to keep the original order of the index for y
	for (int i = 0; i < nSample; i ++) {
		orderPermuted.push_back(i);
	}
	
	// update the observed statistics for each gene
	for (int i = 0; i < nGenes; i ++) {
		// update the xEachGene
		RVanalysisUpdateEachGene(xAllGenes, i);
		eachPermutation.Initializer(xEachGene,xCovIn, yObs);
		// regress on the observed phenotypes
		eachPermutation.RVregPermuteObsWithCov(choice,userWeight);
		// update the observed statitics
		statObs[i] = eachPermutation.statistics;
		// update the p values -- this is useful for the genes with uniform rv scores across all subjects
		pFinalEachGene[i] = eachPermutation.p; 
		// update the same score indicator
		eachGeneSameScore[i] = eachPermutation.sameScore;
		// for those that desires permuation (!sameScore) , we should intialize their p values to be 0
		if (!eachGeneSameScore[i]) {
			pFinalEachGene[i] = 0;
		}
	}
	
	// permute y and each time, update those that does not have the same rare variant scores
	for (int i = 0; i < nPermutation; i ++) {
		// shuffle the order
		gsl_ran_shuffle(g_r, &(orderPermuted[0]), nSample, sizeof(int));
		
		// update the phenotype at this permutation order
		for (int j = 0; j < nSample; j ++) { // update the permuted phenotype
			yPermuted[j] = yObs[orderPermuted[j]];
		}
		
		// go through all genes
		for (int l = 0; l < nGenes ; l++) { // using the same phentoype for each gene
			if (!eachGeneSameScore[l]) {
				// updat the xEachGene
				RVanalysisUpdateEachGene(xAllGenes, l);
				// need to update the memembers in eachPermutation
				eachPermutation.Initializer(xEachGene,xCovIn,yPermuted,1); 
				eachPermutation.RVregPermuteObs(choice,userWeight); // Note: Don't use RVPermuteEach() since the xGenoCov needs to be updated here every time
				statEachPermute[l] = eachPermutation.statistics;
				pGenesEachPermute[l] = eachPermutation.p;
				// update the p values at this step 
				PvaluesUpdateStep(l,statEachPermute[l],statObs[l]); 
			}
		}
		
		// sort the p's
		pGenesEachPermute.Sort();
		pQuantilesPermutation[i][0] = pGenesEachPermute[idxLower];
		pQuantilesPermutation[i][1] = pGenesEachPermute[idxUpper];	
	}
	
  // after the permutation is done, need to finalize the p values
	PvaluesFinalizeStep();
}


Matrix & RVanalysisGlobalPermutation::GetPs(){
	for (int i = 0; i < nPermutation; i ++) {
		for (int j = 0; j < 2; j ++) {
			printf("pQuantilesPermutation[%d][%d] = %f \t", i, j, pQuantilesPermutation[i][j]);
		}
		printf("\n");
	}
	return pQuantilesPermutation;
}


//// A bunch of easy printing tools
//// 
void RVanalysisGlobalPermutation::PrintEachGeneInfo(int i)
{
	// print out the total rare variant counts, number of cases with rv, number of ctrls with rv
	 printf("%d, %s, (%d,%d) %d, %d, %d, ", vec_geneInfo[i].basicGeneInfo.iChromosome,vec_geneInfo[i].basicGeneInfo.strName.c_str(), 
					vec_geneInfo[i].basicGeneInfo.iStart, vec_geneInfo[i].basicGeneInfo.iEnd,
					vec_geneInfo[i].basicGeneInfo.iEnd - vec_geneInfo[i].basicGeneInfo.iStart + 1,
					vec_geneInfo[i].numCase1, vec_geneInfo[i].numCtrl1);
}

void RVanalysisGlobalPermutation::PrintVecIndices(){
	for (int i = 0; i < nGenes; i ++) {
		printf("start = %d, end = %d\t", vec_geneInfo[i].basicGeneInfo.iEnd, vec_geneInfo[i].basicGeneInfo.iEnd);
	}
}

void RVanalysisGlobalPermutation::PrintEachGeneSameScore(){
	for (int i = 0; i < nGenes; i ++) {
		PrintEachGeneInfo(i);
		printf("eachGeneSameScore[i] = %d\n", (int) eachGeneSameScore[i]);
	}
}

void RVanalysisGlobalPermutation::PrintPFinals(){
	printf("~~~~~~~~~~~~~~~~~~ p value for each gene from permutation ~~~~~~~~~~~~~~~~~~~~\n");
	for (int i = 0; i < nGenes; i ++) {
		PrintEachGeneInfo(i);
		printf("pFinalEachGene = %f\n", pFinalEachGene[i]);
	}
}

void RVanalysisGlobalPermutation::PrintPAsym(){
	// printf("~~~~~~~~~~~~~~~~~~ p value for each gene from permutation ~~~~~~~~~~~~~~~~~~~~\n");
	for (int i = 0; i < nGenes; i ++) {
		PrintEachGeneInfo(i);
		printf("%f\n",pGenesEachPermute[i]);
	}
}

void RVanalysisGlobalPermutation::PrintCmpPAsymPpermute(){
	for (int i = 0; i < nGenes; i ++) {
		PrintEachGeneInfo(i);
		printf("%f, %f\n",pGenesAsy[i],pFinalEachGene[i]);
	}
}

void RVanalysisGlobalPermutation::WritePs(const String & outPsfileName, bool outPpermute){
	FILE * outPvalue = fopen(outPsfileName, "w");
	if (outPvalue == NULL) {
		printf("Error in opening %s!\n",outPsfileName.c_str());
		return ;
	}
	fprintf(outPvalue, "%s\t%s\t%s\t%s\t%s\t%s\n", "chrm", "geneName", "rvCounts", "numCase1", "numCtrl1","pValue");
	if (outPpermute) { // outpute p values from permutation
		for (int i = 0; i < nGenes; i ++) {
			fprintf(outPvalue,"%d\t %s\t %d\t %d\t %d\t ", vec_geneInfo[i].basicGeneInfo.iChromosome,vec_geneInfo[i].basicGeneInfo.strName.c_str(), 
							vec_geneInfo[i].basicGeneInfo.iEnd - vec_geneInfo[i].basicGeneInfo.iStart + 1,
							vec_geneInfo[i].numCase1, vec_geneInfo[i].numCtrl1);
			fprintf(outPvalue,"%1.5e\n",pFinalEachGene[i]);
		}		
	} else { // output p value from asymptotic distribution
		for (int i = 0; i < nGenes; i ++) {
			fprintf(outPvalue,"%d\t %s\t %d\t %d\t %d\t ", vec_geneInfo[i].basicGeneInfo.iChromosome,vec_geneInfo[i].basicGeneInfo.strName.c_str(), 
							vec_geneInfo[i].basicGeneInfo.iEnd - vec_geneInfo[i].basicGeneInfo.iStart + 1,
							vec_geneInfo[i].numCase1, vec_geneInfo[i].numCtrl1);
			fprintf(outPvalue,"%1.5e\n",pGenesAsy[i]);
		}		
	}
	fclose(outPvalue);
}

void RVanalysisGlobalPermutation::WriteGlobalPermuteSummary(const String & globalPermutationFileName){
	FILE * outPbounds = fopen(globalPermutationFileName, "w");
	if (outPbounds == NULL) {
		printf("Error in opening %s!\n",globalPermutationFileName.c_str());
		return ;
	}
	fprintf(outPbounds, "%s\t%s\n", "fivePercentQuantile","ninetyFivePercentQuantile");
	for (int i = 0; i < nPermutation; i ++) {
		fprintf(outPbounds, "%f, %f\n", pQuantilesPermutation[i][0], pQuantilesPermutation[i][1]);
	}
	fclose(outPbounds);
}