/*
 *  RVanalysisGlobalPermutation.h
 *  LogitReg
 *
 *  Created by Youna Hu on 1/6/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */
#ifndef __RVANALYSIS_GLOBAL_PERMUTATION_H__
#define __RVANALYSIS_GLOBAL_PERMUTATION_H__

#include "Subset012byGene.h"
#include "rareVariantsAnalysis.h"

extern gsl_rng * g_r;

struct GeneOutINFO {
	Gene basicGeneInfo;
	int numCase1;
	int numCtrl1;
};

// The idea is to have a matrix pPermute(nPermute,2) 
// to record the 5% and 95% quantile of the p values across all genes at each permutation
class RVanalysisGlobalPermutation{
public:
	//static gsl_rng *rGSL;
	
	RareVariantAnalysis eachPermutation; // each permutation is an object of rareVariantAnalysis 
																				// note there is a nPermute is used for gene level permutation.
																				// if we have global permutation, we do not need gene level permutation
																				// so the gene level permutation should be defaulted to be 0
	int nPermutation; // the number of permutation used in the global permutation
	int nGenes; // number of genes
	int nSample;
	int nEachGeneSize;
	int nEachGeneStart;
	int idxLower, idxUpper;
	Matrix xEachGene;
	std::vector<int> orderPermuted;
	Vector pGenesEachPermute; // a vector to contain the p values for each gene at each permutation
	Vector pFinalEachGene; // store the p values from permutation
	Vector pGenesAsy; // store the asympototic p values
	Vector statObs; // statistics for observed data for each gene
	Vector statEachPermute; // a vector to contain the statistics for each gene at each permutation, -- then use this to calcualte p values
	Matrix pQuantilesPermutation; // the 5% and 95% quantile of the p value
	Vector yPermuted; // the permuted genotype vector
	std::vector<bool> eachGeneSameScore;
	
	std::vector<GeneOutINFO> vec_geneInfo; // Note: Even we borrow the Gene structure, with iChromosome, strName, iStart, iEnd
																	// this geneInfo vector stores, iChromsome, strName, idx_start, idx_end
	
	// Init
	void Initializer(); // get rid of this one
	void Initializer(RareVariantAnalysis &rvAnalysisBaseValue, int timesPermutation);
	void Initializer(RareVariantAnalysis &rvAnalysisBaseValue, Matrix &xAllGenes, 
									 Vector &yObs, int timesPermutation);
	
	void Initializer(RareVariantAnalysis &rvAnalysisBaseValue, Matrix &xAllGenes, 
									 Vector &yObs, int timesPermutation, 
									 Matrix &positionMatrix, 
									 const std::string &geneListFileName);
	
	// constructor
	RVanalysisGlobalPermutation();
	RVanalysisGlobalPermutation(RareVariantAnalysis &rvAnalysisBaseValue, int timesPermutation);
	RVanalysisGlobalPermutation(RareVariantAnalysis &rvAnalysisBaseValue, Matrix &xAllGenes, 
															Vector &yObs, int timesPermutation);
	RVanalysisGlobalPermutation(RareVariantAnalysis &rvAnalysisBaseValue, Matrix &xAllGenes, 
															Vector &yObs, int timesPermutation, 
															Matrix &positionMatrix, const std::string &geneListFileName);
	
	// global permutation
	void RVanalysisUpdateGeneInfo(Matrix &positionMatrix, const std::string &geneListFileName); // update the index for each gene
	
	void RVanalysisUpdateEachGene(Matrix & xAllGenes, int idxGeneIn); // update xEachGene to be the ith gene
	
	// the global permuation() function only record the pQuantilesPermuation for a QQplot
	void GlobalPermutation(Matrix &xAllGenes, Matrix &positionMatrix, const std::string & geneListFileName, 
												 Vector &yObs, std::string choice, Vector &userWeight);
	// get the asymptotic p values for each gene -- stored in pGeneEachPermute
	void PvaluesGeneAsy(Matrix &xAllGenes, Vector &yObs, std::string choice, Vector &userWeight); // note: need to write exact logistic regression
	void PvaluesGeneAsyWithCov(Matrix &xAllGenes, Matrix &xCovIn, Vector &yObs, std::string choice, Vector &userWeight);
	
	// The by-product of a global permutation is the p value for each gene calcualted by permutation
	// since we decide not to save the statistics at each permutation for each gene, 
	// we need the following two functions for final p value calculation -- pValueUpdateStep is to update the pValue at each permutation, 
	void PvaluesUpdateStep(int i, double statPermuted,double statObserved);
	void PvaluesFinalizeStep(); 
	// Two products: 1. get the 5% and 95% quantile of the p values for each gene at each step 2. the final p value for each gene using the permutation approach
	void PvaluesGenePermutation(Matrix &xAllGenes, Vector &yObs, std::string choice, Vector &userWeight);
	void PvaluesGenePermutationWithCov(Matrix &xAllGenes, Matrix &xCovIn, Vector &yObs, std::string choice, Vector &userWeight);
	
	// getters
	Matrix & GetPs();
	void PrintEachGeneInfo(int i);
	void PrintVecIndices();
	void PrintEachGeneSameScore();
	void PrintPFinals();
	void PrintPAsym();
	void PrintCmpPAsymPpermute();
	
	
	// write p values to file 
	void WritePs(const String & outPsfileName, bool outPpermute);
	void WriteGlobalPermuteSummary(const String & globalPermutationFileName);	
	// desctructor
	~RVanalysisGlobalPermutation();
};

#endif