/*
 *  Subset012.cpp
 *  LogitReg
 *
 *  Created by Youna Hu on 1/2/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include "Subset012byGene.h"
#include "MathMatrix.h"
// constructor of the positionFile
PositionFile::PositionFile(const std::string & strFile) {
	std::ifstream fsPositions;
	fsPositions.open(strFile.c_str(), std::ios::in);
	if (!fsPositions.is_open()) {
		std::cout << "Unable to open position file " << strFile << std::endl;
		throw;
	}
	
	for (;;) {
		int iChromosome;
		int iPosition;
		
		std::string chrBuffer;
		fsPositions >> chrBuffer;
		
		// check if file has ended
		if (fsPositions.eof()) {
			break;
		}		
		if (chrBuffer == "X" || chrBuffer == "x"){
			iChromosome = 23;
		} else if (chrBuffer == "Y" || chrBuffer == "y"){
			iChromosome = 24;
		} else{
			iChromosome = atoi(chrBuffer.c_str());
		}
		
		if (iChromosome > 24 || iChromosome <= 0){
			error("Invalide chromosone numbers.\n");
		}
		fsPositions >> iPosition;
		
		if (iChromosome > 10000) {
			std::cout << "Sanity check failed: Chromosome number too large." << std::endl;
			throw;
		}
		// m_vecChromosome size should be the number of chromosones in consideration
		if (iChromosome >= m_vecChromosomes.size()) {
			m_vecChromosomes.resize(iChromosome+1, NULL);
		}
		
		if (m_vecChromosomes[iChromosome] == NULL) {
			m_vecChromosomes[iChromosome] = new PositionVector;
		}
		
		m_vecChromosomes[iChromosome]->push_back(iPosition);
	}
	
	// Output number of lines read
	std::cout << "Found " << m_vecChromosomes.size()-1;
	std::cout << " chromosone(s)." << std::endl;
}

PositionFile::PositionFile(const Matrix & posMatrix){
	
	if(posMatrix == NULL) 
		error("Empty position Matrix\n");
	// update the m_vecChromosome
	for (int i = 0;i < posMatrix.rows; i++){
		int iChromosome = (int) posMatrix[i][0];
		int iPosition = (int) posMatrix[i][1];
		
		if (iChromosome > 24){
			error("Sanity check failed: Chromosome number too large\n");
		}
		
		// m_vecChromosome size should be the number of chromosones in consideration
		if (iChromosome >= m_vecChromosomes.size()) {
			m_vecChromosomes.resize(iChromosome+1, NULL);
		}
		
		if (m_vecChromosomes[iChromosome] == NULL) {
			m_vecChromosomes[iChromosome] = new PositionVector;
		}
		
		m_vecChromosomes[iChromosome]->push_back(iPosition);		
	}
	// Output number of lines read
	std::cout << "Found " << m_vecChromosomes.size()-1;
	std::cout << " chromosome(s)." << std::endl;
}

// Destructor
PositionFile::~PositionFile() {
	int i;
	
	for (i = 0; i < m_vecChromosomes.size(); i++) {
		if (m_vecChromosomes[i] != NULL) {
			delete m_vecChromosomes[i];
		}
	}
}

// Find a position in the file
int PositionFile::FindElement(
								int iChromosome,
								int iPosition,
								bool fEnd
								) {
	
	// Check that this chromosone is defined
	if (iChromosome >= m_vecChromosomes.size()) {
		std::cout << "Chromosome index too large." << std::endl;
		std::cout << "iChromosome = " << iChromosome << std::endl;
		return SEARCH_FAIL;
	}
	
	if (m_vecChromosomes[iChromosome] == NULL) {
		std::cout << "No information available on chromosone "
		<< iChromosome << std::endl;
		return SEARCH_FAIL;
	}
	
	// Reference to the position vector
	const PositionVector & vecPosition =
	*(m_vecChromosomes[iChromosome]);
	
	// Use binary search to find the start
	int iLeft = 0;
	int iRight = vecPosition.size() - 1;
	
	int iPos;
	
	if (vecPosition[iRight] < iPosition) {
		return TOO_BIG;
	}
	if (vecPosition[iLeft] > iPosition) {
		return TOO_SMALL;
	}
	
	for (;;) {
		iPos = (iLeft + iRight) / 2;
		
		if (vecPosition[iPos] == iPosition) {
			break;
		} else if (vecPosition[iPos] < iPosition) {
			if (iLeft == iPos) {
				iPos = iRight;
				break;
			}
			iLeft = iPos;
			
		} else {
			// this situation will never happen
			if (iRight == iPos) {
				throw;
			}
			iRight = iPos;
		}
	}
	
	if ((vecPosition[iPos] != iPosition) && (fEnd)) {
		iPos--;
	}
	
	return iPos;
}

// Find a range within the file using binary search
bool PositionFile::FindRange(
							 int iChromosome,
							 int iStart,
							 int iEnd,
							 int & iStartPos,
							 int & iEndPos
							 ) {
	
	// Check that this chromosone is defined
	if (m_vecChromosomes[iChromosome] == NULL) {
		std::cout << "No information available on chromosone "
		<< iChromosome << std::endl;
		return false;
	}
	
	// Find the start position
	iStartPos = FindElement(iChromosome, iStart, false);
	
	// Find the end position
	iEndPos = FindElement(iChromosome, iEnd, true);
	
	// Reference to the position vector
	const PositionVector & vecPosition =
	*(m_vecChromosomes[iChromosome]);
	
	// Check range
	if ((iStartPos == SEARCH_FAIL) || (iEndPos == SEARCH_FAIL)) {
		iStartPos = SEARCH_FAIL;
		iEndPos = SEARCH_FAIL;
		return false;
	}
	
	if ((iStartPos == TOO_SMALL) && (iEndPos == TOO_BIG)) {
		iStartPos = 0;
		iEndPos = vecPosition.size()-1;
	} else if ((iStartPos == TOO_SMALL) && (iEndPos != TOO_BIG)){
		iStartPos = 0;
	} else if ((iStartPos != TOO_SMALL) && (iEndPos == TOO_BIG)){
		iEndPos = vecPosition.size()-1;
	}
	
	// if iStartPos is too big then, search failed
	if (iStartPos == TOO_BIG || iEndPos == TOO_SMALL) {
		return false;
	}
	
	if (iEndPos < iStartPos) {
		iStartPos = NOTHING_IN_RANGE;
		iEndPos = NOTHING_IN_RANGE;
		return false;
	}
	
	// When iStartPos = iEndPos, we would have the choosen position is at the bounary
	
	
	// Since the iStartPos and iEndPos are the positions at the position vector for each chromosome,
	// We will need to adjust iStartpos and iEnd here
	int offset = 0; // do I need to declare offset as long integer?
	if (iChromosome == 1){
		offset = 0;
	} else {
		for (int i = 1; i < iChromosome; i ++) {
			PositionVector & vecPositionTMP = *(m_vecChromosomes[i]);
			//					printf("Chromosome %d has %d variants\n",i,(int) vecPositionTMP.size());
			offset += vecPositionTMP.size();
		}
	} 
	
	//			printf("offset = %d\n", offset); 
	// So m_vecChromosomes is designed to not use m_vecChromosomes[0] (= NULL)
	
	// adjust the positions
	iStartPos += offset;
	iEndPos += offset;
	
	return true;
}


GeneFile::GeneFile(const std::string & strFile) {
	
	std::ifstream fsGenes;
	fsGenes.open(strFile.c_str(), std::ios::in);
	if (!fsGenes.is_open()) {
		error("Unable to open gene file ");
	}
	
	for (;;) {
		Gene geneData;
		
		std::string strBuffer;
		fsGenes >> strBuffer;
		
		if (fsGenes.eof()) {
			break;
		}
		strBuffer = strBuffer.substr(0,strBuffer.length()); 
		
		if (strBuffer == "X" || strBuffer == "x"){
			geneData.iChromosome = 23;
		} else if (strBuffer == "Y" || strBuffer == "y"){
			geneData.iChromosome = 24;
		} else{
			geneData.iChromosome = atoi(strBuffer.c_str());
		}
		
		// sanity check
		if (geneData.iChromosome > 24 || geneData.iChromosome <= 0){
			error("Invalid chromosome numbers\n");
		}
		
		fsGenes >> geneData.strName;
		fsGenes >> geneData.iStart;
		fsGenes >> geneData.iEnd;
		m_vecGenes.push_back(geneData);
	}
	
	std::cout << "Read " << m_vecGenes.size() << " genes" << std::endl;
}

// Get a gene
const Gene & GeneFile::GetGene(int iGene) const {
	return m_vecGenes[iGene];
}

// Number of genes
size_t GeneFile::GetGeneCount() const {
	return m_vecGenes.size();
}

// get the index of a given gene
// right now, just use linear search, can upgrade it to binary search later
int GeneFile::SearchGene(String &geneNameIn) const{
	std::string std_geneName(geneNameIn);
	int i;
	for (i = 0; i < m_vecGenes.size(); i ++) {
		if (m_vecGenes[i].strName == std_geneName){
			return (i);	
		}
	}
	if (i == m_vecGenes.size()){
		printf("Gene %s does not exist!\n",geneNameIn.c_str());
		return (-1);
	}	
}
