/*
 *  Subset012sbyMAF.cpp
 *  LogitReg
 *
 *  Created by Youna Hu on 1/15/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#include "Subset012sbyMAF.h"

void Subset012sbyMAF::CHECKprintMat(Matrix &x){
	for (int i = 0; i < x.rows; i ++) {
		for (int j = 0; j < x.cols; j ++) {
			printf("%f\t",x[i][j]);
		}
		printf("\n");
	}	
}

void Subset012sbyMAF::CHECKprintVec(Vector &x){
	for (int i = 0; i < x.Length(); i++) {
		printf("%f\n",x[i]);
	}	
}

void Subset012sbyMAF::ColumnSum(Matrix & genoMat, Vector & af){
	//	Vector af;
	af.Dimension(genoMat.cols,0);
	for (int i = 0; i < genoMat.cols; i ++) {
		for (int j = 0; j < genoMat.rows; j ++) {
			af[i] += genoMat[j][i];
		}
	}
	//	return (af);
}

bool Subset012sbyMAF::CheckGenoValid(Matrix &xGeno){
	for (int i = 0; i < xGeno.rows; i++) {
		for (int j = 0; j < xGeno.cols; j++) {
			if (xGeno[i][j] != 0 && xGeno[i][j] != 1 && xGeno[i][j] != 2) {
				printf("Invalid genotype (%1.0f) at element (%d,%d)!\n",xGeno[i][j],i,j);
				return(false);
			}
		}
	}
	return(true);
}

bool Subset012sbyMAF::CheckPhenoValid(Vector &yPheno){
	for (int i = 0; i < yPheno.Length(); i++) {
		if (yPheno[i] != 0 && yPheno[i] != 1) {
			printf("Invalid phenotype (%1.0f) at element (%d)!\n",yPheno[i],i);
			return(false);
		}
	}
	return(true);
}

void Subset012sbyMAF::FlipGenoReference(Matrix & genoMat){
	Vector genoAF;
	ColumnSum(genoMat,genoAF);
	int nSampleSize = genoMat.rows;
	// if the genoAF is actually the MAF, otherwise, change the reference allele
	for (int j = 0; j < genoAF.Length(); j++) {
		if (genoAF[j] > 0.5*nSampleSize) {
			for (int i = 0; i < nSampleSize; i++) {
				genoMat[i][j] = 2-genoMat[i][j];
			}
		}
	}
}

void Subset012sbyMAF::GenoPosSubsetbyMAF(Matrix & posMat, Matrix & genoMat, String & frqFileName, double mafCutoff)
{
	std::vector<double> frqVector;	
	int numRareVariants = 0;
	// store the 5th row of frqFileName as
	FILE *frqFileIn = fopen(frqFileName, "r");
	if (frqFileIn == NULL) {
		printf("Error opening file (%s)!\n",frqFileName.c_str());
		return;
	}
	String lineEntry;
	StringArray lineEntryArray;
	int refAFidx = 4;
	// this one is to read the header file but doesn't do anything.
	lineEntry.ReadLine(frqFileIn);
	
	while (!feof(frqFileIn)){
		lineEntry.ReadLine(frqFileIn);
		lineEntryArray.AddTokens(lineEntry, "\t");
		double	refFrq = lineEntryArray[refAFidx].LeftClip(2).AsDouble();
		if (refFrq <= mafCutoff || refFrq >= 1-mafCutoff){
			numRareVariants = numRareVariants + 1;
		}
		frqVector.push_back(refFrq);
		lineEntry.Clear();
		lineEntryArray.Clear();
	}	
	fclose(frqFileIn);
	// sanity check - the number of column for the genoMat should be the same as that for the frq File
	if (genoMat.cols	!= frqVector.size()){
		printf("Error: The column size of 012 matrix is not the same as the row size of the frq File.\n");
		return;
	}
	// Now, we need to copy the rare varaint positions to a new Matrix
	int totalPosCount = frqVector.size(); // number of positions
	int totalSbjCount = genoMat.rows; // number of subjects
	Matrix genoMatRare(totalSbjCount,numRareVariants);
	Matrix posMatRare(numRareVariants,2);
	
	int idxRare = 0; // index to increment the rare varaint genoMat and posMat
	// update the new matrices (genoMat and posMat)
	for (int i = 0; i < totalPosCount; i ++) {
		if (frqVector[i] <= mafCutoff || frqVector[i] >= 1-mafCutoff) {
			//*(genoMatRare[idxRare]) = genoMat[i]; // this does not work
			// copy the column of rare variants in genoMat to genoMatRare
			for (int j = 0; j < totalSbjCount; j ++) {
				genoMatRare[j][idxRare] = genoMat[j][i];
			}
			// copy the row of rare variants in posMat to posMatRare
			posMatRare[idxRare][0] = posMat[i][0]; //copy the chromosome number
			posMatRare[idxRare][1] = posMat[i][1]; // copy the position number
			idxRare = idxRare + 1;			
		}
	}
	// Copy the subsetted matrix into the original matrices
	genoMat = genoMatRare;
	posMat = posMatRare;
}