/*
 *  rareVariantsAnalysis.h
 *  LogitReg
 *
 *  Created by Youna Hu on 1/3/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */
#ifndef __RAREVARIANTANALYSIS_H__
#define __RAREVARIANTANALYSIS_H__

#include "LogisticRegression.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_sf_gamma.h>
#include <string>
#include <vector>
#include "Table2by2.h"
#include "MathVector.h"
#include "MathMatrix.h"
#include "Error.h"

extern gsl_rng * g_r;

class RareVariantAnalysis{
public:
	double statistics;
  double p;
	int nSample;
	int nVariants;
	
	/*** covariates: intercept, genotype, collapsed genotype, covariates ***/
	Matrix xIntercept;
	Matrix xCov;
	Matrix xInterceptCov; // columns: (intercept, xCov) this is used for likelihood ratio test
	Matrix xGeno;	
	Vector xGenoVariant; // genotype for each variant, used in CalcFreqCtrl() 
	Vector	rvWeight; // weight for each rare variant
	Matrix xGenoCombinedMat; // this one get updated differentially according to specified methods
	Matrix xGenoCov; // covariate matrix, row - subject, column - covariates: (intercept, xGenoCombinedMat, xCov)
	
	/*** phenotype variables ***/
	Vector yPhenoReg; // phenotype in Vector data form to be used in logistic regression
	Vector yControl1;
	
	bool sameScore;
	Table2by2 twoBytwoMat;
	
	/*** permutation variables ***/
	int nPermute;
	std::vector<int> orderPermute; // newly shuffled order
	Vector yPhenoPermute; // length nSample
	Vector statPermute; // lenth nPermute
	Vector pPermute;
	double statObserved;
	
	// Initializer or called resetter 
	void Initializer(Matrix & xGene, Vector & yIn, int nPermutation= 0);
	void Initializer(Matrix &xGene, Vector &yIn, std::string &choiceCollapse, int nPermutation = 0);
	void Initializer(Matrix &xGene, Matrix &xCovIn, Vector &yIn, int nPermutation = 0);
	void Initializer(Matrix &xGene, Matrix &xCovIn, Vector &yIn, std::string &choiceCollapse, int nPermutation = 0);
	
	// constructor
	RareVariantAnalysis();
	RareVariantAnalysis(Matrix & xGene, Vector & yIn, int nPermutation= 0);
	RareVariantAnalysis(Matrix & xGene, Vector & yIn, std::string &choiceCollapse, int nPermutation= 0);
	RareVariantAnalysis(Matrix &xGene, Matrix &xCovIn, Vector &yIn, int nPermutation = 0);
	RareVariantAnalysis(Matrix &xGene, Matrix &xCovIn, Vector &yIn, std::string & choiceCollapse, int nPermutation = 0);
	
	// destructor 
	~RareVariantAnalysis();
	
	// rare variant weights for calculating rare variants scores
	void CalcFreqCtrl(bool smallSampleAdjusts = true);
	void WeightMadsonBrowning(); // Update the Madson-Browing weight
	void SetWeight(Vector & userWeight); // Update with a user specified weight
	
	// collapse xGenoCombined
	void CollapseType(std::string choice, Vector &userWeight); // Collapsing choice can be "or","sum","wt","user"
	
	// RV methods --- with only rare variant scores as covariates
	// Using regression
	void RVregression(std::string choice, Vector &userWeight); 
	void RVregressionWithCov(std::string choice, Vector &userWeight); // ?????? -- need to write exact logistic regression
	
	// Permutation
	void Pvalue(bool twoSided = true, bool smallerIsExtremer = true);
	void RVregPermuteObs(std::string choice, Vector &userWeight); 
	void RVregPermuteObsWithCov(std::string choice, Vector &userWeight);
	
	void RVregPermuteEach(std::string choice, Vector &userWeight); // Output the deviance for the purpose of permutation, remember here the xGenoCov has been updated
	void RVregPermute(std::string choice, Vector &userWeight);
	
	
	// RV methods --- include covariates
	// ??? to do
	
	
	// Getter functions
	Matrix & GetGenoCombined();
	Vector & GetWeights();
	double &GetP();
	double &GetStatistics();
};

#endif
